﻿package ole.raeumlicheStatistik;

import Geometry2D;

import java.util.*;

public class RSBlatt5 {

	static final Geometry2D.Point zero = new Geometry2D.Point(0,0);
	static final Random r = new Random();
	
	
/* *****************************************
 *  Aufgabe 1                              *
 *******************************************/
	
	public static double distance(double x_1, double y_1, double x_2, double y_2)
	{
		return Math.sqrt((x_1-x_2)*(x_1-x_2) + (y_1-y_2)*(y_1-y_2));		
	}
	
	public static Vector realisePVT(double[] x_axis, double[] y_axis, double intensity)
	{
		// Simuliere einen Poisson-Prozess in einem 
		// genuegend grossen Fenster
		Vector points = PoissonProcess.PoissonProcess(x_axis, y_axis, new PoissonProcess.HomogeneousIntensityFunction(intensity));
		Vector cells = new Vector(points.size());
		
		Geometry2D.Point[] window_corners = new Geometry2D.Point[] {new Geometry2D.Point(x_axis[0], y_axis[0]), new Geometry2D.Point(x_axis[1],y_axis[0]), new Geometry2D.Point(x_axis[1], y_axis[1]), new Geometry2D.Point(x_axis[0], y_axis[1])}; 
		Geometry2D.ConvexPolygon window = new Geometry2D.ConvexPolygon(window_corners);
		
		// Betrachte dann fuer jeden Punkt das gesamte Fenster
		// als zugehoerige Ausgangszelle
		for(int i=0; i<points.size(); i++)
		{
			cells.add(i, window);			
		}
		
		// Schneide fuer jeden Punkt die jeweilige Zelle mit der
		// Halbebene, die durch die jeweiligen Bisektoren der 
		// Verbindungsstrecke mit allen anderen Punkten definiert wird
		for(int i=0; i<points.size(); i++)
		{
	        for(int j=0; j<points.size(); j++)
	        {
	        	if(i!=j)
	        	{
	        	   double[] left_point = (double[]) points.get(i);
	        	   double[] right_point = (double[]) points.get(j);
	        	
	               Geometry2D.Line l = Geometry2D.bisector(new Geometry2D.Point(left_point[0],left_point[1]), new Geometry2D.Point(right_point[0], right_point[1])); 	   
	               Geometry2D.ConvexPolygon cell = (Geometry2D.ConvexPolygon) cells.remove(i);
	         	   cell = Geometry2D.intersectWithHalfplane(cell, l);
	         	   cells.add(i,cell);
	         	}
	        }
		}
	   // Dies liefert für jeden Punkt die Voronoi-Zelle
	   return cells;
	}
	
    // Finde fuer eine Tessellation diejenige Zelle, 
	// welche die Null enthaelt.
	public static Geometry2D.ConvexPolygon getZeroCell(Vector cells) throws Exception
	{
		for(int i=0; i<cells.size(); i++)
		{
			Geometry2D.ConvexPolygon cell = (Geometry2D.ConvexPolygon) cells.get(i);			
			if(cell.contains(zero))
			{
				return cell;
			}
		}
		
		throw new Exception("Could not find zero in this tessellation!");
	}
	
	// Bestimme, ob ein Randeffekt bei der Realisierung der Null-Zelle aufgrund der 
	// Beschraenktheit des Beobachtungsfensters eine Rolle gespielt haben kann
	public static int EdgeEffectPossible(Geometry2D.ConvexPolygon cell, double[] x_axis, double[] y_axis)
	{
		// Bestimme den Abstand des Null-Punktes zum Rand des Beobachtungsfensters
		double distance_to_edge = Math.min( distance(0,0,x_axis[0],0), distance(0,0,x_axis[1],0));
		distance_to_edge = Math.min(distance_to_edge, distance(0,0,0,y_axis[0]));
		distance_to_edge = Math.min(distance_to_edge, distance(0,0,0,y_axis[1]));
		
		//Geometry2D.Point[] vertex = cell.p;
		double distance_to_vertex = findMaxCornerDistance(cell);//= Double.NEGATIVE_INFINITY;
		
		// Wenn der Abstand des Null-Punktes zum Rand geringer ist, als seine doppelte
		// Entfernung von der weitestmoeglich entfernten Ecke der Null-Zelle, dann haette
		// die Zelle evtl. noch durch eine Erweiterung des Beobachtungsfensters veraendert
		// werden koennen, d.h. ein Randeffekt waere moeglich.
		// Ansonsten sind keine Randeffekte moeglich, da die Punkte ausserhalb des Fensters
		// zu weit entfernt sind, um die Null-Zelle noch zu beeinflussen.
		if(distance_to_edge<2*distance_to_vertex)
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}
	
	public static double[] getCellCharacteristics(Geometry2D.ConvexPolygon cell, double[] x_axis, double[] y_axis) //throws Exception
	{
		double[] return_values = {0,0,0,0};
		
		return_values[0] = cell.p.length;
		return_values[1] = cell.getPerimeter();
		return_values[2] = cell.getArea();
		return_values[3] = EdgeEffectPossible(cell, x_axis, y_axis);
		
		return return_values;
	}
	
	//Bestimme den maximalen Abstand einer Ecke 
	//der uebergebenen Zelle vom Null-Punkt
	public static double findMaxCornerDistance(Geometry2D.ConvexPolygon poly)
	{
		double max = Double.NEGATIVE_INFINITY;

		for(int i=0; i<poly.p.length; i++)
		{
		    double d = poly.p[i].distanceTo(zero);
			if(d>max) max = d;
		}

	  return max;
	}

	public static double[] Aufgabe1c()
	{
		double[] mean_values = {0,0,0,0};
		double[] sum_of_squares = {0,0,0,0};
		double[] sd = {0,0,0,0};

		try{
			  // Betrachte das vorgeschlagene Beobachtungsfenster,
			  double[] x_axis = new double[]{-250,250};
			  double[] y_axis = new double[]{-250,250};
			  
			  // ... die vorgegebene Anzahl von Durchlaeufen... 
			  // ... mit der vorgegebenen Intensitaet.
			  final int number_of_runs = 100000;
			  final double intensity = 0.0005;
			   
			  // Fuer jeden Durchlauf...
			  for(int i=0; i<number_of_runs; i++)
			  {
				  // ...realisiere eine PVT und finde ihre Null-Zelle...
				  Vector pvt = realisePVT(x_axis, y_axis, intensity);
				  Geometry2D.ConvexPolygon zero_cell = getZeroCell(pvt);
				  
				  // ...bestimme deren Charakteristiken ...
				  double[] result = getCellCharacteristics(zero_cell, x_axis, y_axis);
				  
				  // ... und berechne so Mittelwert und Standardabweichung.
				  for(int j=0; j<4; j++)
				  {
					  mean_values[j] += result[j];
					  sum_of_squares[j] += result[j] * result[j];
				  }
			  }
			  
			  for(int j=0; j<3; j++)
			  {
				  mean_values[j] /= number_of_runs;
				  sd[j] = Math.sqrt((sum_of_squares[j] - number_of_runs * mean_values[j] * mean_values[j])/(number_of_runs-1));
			  }
			  
			  // Benutze die bestimmten Werte dann, um asymptotische Tests 
			  // auf die hypothetischen Werte der TYPISCHEN Zelle durchzufuehren. 
			  System.out.println("Mean Number Of Vertices: "+mean_values[0]);
			  System.out.println("Test for Expectation=6:");
			  System.out.println("--------------------------------------------------");
			  
			  System.out.println("Mean Perimeter: "+mean_values[1]);
			  System.out.println("Test for Expectation="+(4 / Math.sqrt(intensity))+":");
			  System.out.println("--------------------------------------------------");
			 
			  System.out.println("Mean Area: "+mean_values[2]);
			  System.out.println("Test for Expectation="+(1 / intensity)+":");
			  System.out.println("--------------------------------------------------");
			 
			  System.out.println("Edge Effects possible in "+mean_values[3]+" of "+number_of_runs+" cases.");
		   }
		   catch(Exception e){ System.err.println("RSBlatt5.Aufgabe1c: "+e); } 
		   
		   return mean_values;
	}
	
	
}